# -*- coding: utf-8 -*-
"""Tests for loading the compressed generic TSV file with custom fields
"""

import io
import json
import textwrap

import pytest

from biomedsheets import io_tsv

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


@pytest.fixture
def tsv_sheet_generic_header():
    f = io.StringIO(textwrap.dedent("""
    [Metadata]
    schema\tgeneric
    schema_version\tv1
    title\tExample generic study
    description\tSimple example of a generic study sample sheet.

    [Custom Fields]
    key\tannotatedEntity\tdocs\ttype\tminimum\tmaximum\tunit\tchoices\tpattern
    ncbiTaxon\tbioEntity\tNCBI taxon ID\tpattern\t.\t.\t.\t.\t^NCBITaxon_[1-9][0-9]*$
    uberonCellSource\tbioSample\tUberon cell source\tpattern\t.\t.\t.\t.\t^UBERON:[0-9]+$
    testSampleConcentration\ttestSample\tConcentration in test sample\tnumber\t.\t.\tug/ml\t.\t.
    ngsLibraryConcentration\tngsLibrary\tConcentration in NGS library\tnumber\t.\t.\tug/ml\t.\t.

    [Data]
    bioEntity\tbioSample\ttestSample\tngsLibrary\textractionType\tlibraryType\tfolderName\tncbiTaxon\tuberonCellSource\ttestSampleConcentration\tngsLibraryConcentration
    E001\tBS1\tTS1\tLIB1\tRNA\ttotal_RNA_seq\tE001-BS1-TS1-LIB1\tNCBITaxon_9606\tUBERON:0002107\t0.03\t0.001
    E001\tBS2\tTS1\tLIB1\tRNA\ttotal_RNA_seq\tE001-BS2-TS1-LIB1\tNCBITaxon_9606\tUBERON:0002107\t0.05\t0.002
    E002\tBS1\tTS1\tLIB1\tRNA\ttotal_RNA_seq\tE002-BS1-TS1-LIB1\tNCBITaxon_9606\tUBERON:0002107\t0.06\t0.004
    E002\tBS1\tTS1\tLIB2\tRNA\ttotal_RNA_seq\tE001-BS1-TS1-LIB2\tNCBITaxon_9606\tUBERON:0002107\t0.08\t0.005
    """.lstrip()))
    return f


@pytest.fixture
def tsv_sheet_generic_no_header():
    """Generic TSV sheet without header"""
    f = io.StringIO(textwrap.dedent("""
    bioEntity\tbioSample\ttestSample\tngsLibrary\textractionType\tlibraryType\tfolderName\tncbiTaxon\tuberonCellSource\ttestSampleConcentration\tngsLibraryConcentration
    E001\tBS1\tTS1\tLIB1\tRNA\ttotal_RNA_seq\tE001-BS1-TS1-LIB1\tNCBITaxon_9606\tUBERON:0002107\t0.03\t0.001
    E001\tBS2\tTS1\tLIB1\tRNA\ttotal_RNA_seq\tE001-BS2-TS1-LIB1\tNCBITaxon_9606\tUBERON:0002107\t0.05\t0.002
    E002\tBS1\tTS1\tLIB1\tRNA\ttotal_RNA_seq\tE002-BS1-TS1-LIB1\tNCBITaxon_9606\tUBERON:0002107\t0.06\t0.004
    E002\tBS1\tTS1\tLIB2\tRNA\ttotal_RNA_seq\tE001-BS1-TS1-LIB2\tNCBITaxon_9606\tUBERON:0002107\t0.08\t0.005
    """.lstrip()))
    return f


# Expected value for the generic sheet JSON with header
EXPECTED_GENERIC_SHEET_JSON_HEADER = r"""
{
    "identifier": "file://<unknown>",
    "title": "Example generic study",
    "description": "Simple example of a generic study sample sheet.",
    "extraInfoDefs": {
        "bioEntity": {
            "ncbiTaxon": {
                "docs": "NCBI taxon ID",
                "key": "ncbiTaxon",
                "type": "pattern",
                "pattern": "^NCBITaxon_[1-9][0-9]*$"
            }
        },
        "bioSample": {
            "uberonCellSource": {
                "docs": "Uberon cell source",
                "key": "uberonCellSource",
                "type": "pattern",
                "pattern": "^UBERON:[0-9]+$"
            }
        },
        "testSample": {
            "extractionType": {
                "docs": "Describes extracted",
                "key": "extractionType",
                "type": "enum",
                "choices": [
                    "DNA",
                    "RNA",
                    "other"
                ]
            },
            "testSampleConcentration": {
                "docs": "Concentration in test sample",
                "key": "testSampleConcentration",
                "type": "number",
                "unit": "ug/ml"
            }
        },
        "ngsLibrary": {
            "seqPlatform": {
                "docs": "Sequencing platform used",
                "key": "kitName",
                "type": "enum",
                "choices": [
                    "Illumina",
                    "PacBio",
                    "other"
                ]
            },
            "libraryType": {
                "docs": "Rough classificiation of the library type",
                "key": "libraryType",
                "type": "enum",
                "choices": [
                    "Panel-seq",
                    "WES",
                    "WGS",
                    "mRNA-seq",
                    "tRNA-seq",
                    "other"
                ]
            },
            "folderName": {
                "docs": "Name of folder with FASTQ files",
                "key": "folderName",
                "type": "string"
            },
            "ngsLibraryConcentration": {
                "docs": "Concentration in NGS library",
                "key": "ngsLibraryConcentration",
                "type": "number",
                "unit": "ug/ml"
            }
        }
    },
    "bioEntities": {
        "E001": {
            "pk": 1,
            "extraInfo": {
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {
                "BS1": {
                    "pk": 2,
                    "extraInfo": {
                        "uberonCellSource": "UBERON:0002107"
                    },
                    "testSamples": {
                        "TS1": {
                            "pk": 3,
                            "extraInfo": {
                                "extractionType": "RNA",
                                "testSampleConcentration": 0.03
                            },
                            "ngsLibraries": {
                                "LIB1": {
                                    "pk": 4,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "E001-BS1-TS1-LIB1",
                                        "libraryType": "total_RNA_seq",
                                        "ngsLibraryConcentration": 0.001
                                    }
                                }
                            }
                        }
                    }
                },
                "BS2": {
                    "pk": 5,
                    "extraInfo": {
                        "uberonCellSource": "UBERON:0002107"
                    },
                    "testSamples": {
                        "TS1": {
                            "pk": 6,
                            "extraInfo": {
                                "extractionType": "RNA",
                                "testSampleConcentration": 0.05
                            },
                            "ngsLibraries": {
                                "LIB1": {
                                    "pk": 7,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "E001-BS2-TS1-LIB1",
                                        "libraryType": "total_RNA_seq",
                                        "ngsLibraryConcentration": 0.002
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        "E002": {
            "pk": 8,
            "extraInfo": {
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {
                "BS1": {
                    "pk": 9,
                    "extraInfo": {
                        "uberonCellSource": "UBERON:0002107"
                    },
                    "testSamples": {
                        "TS1": {
                            "pk": 12,
                            "extraInfo": {
                                "extractionType": "RNA",
                                "testSampleConcentration": 0.08
                            },
                            "ngsLibraries": {
                                "LIB2": {
                                    "pk": 13,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "E001-BS1-TS1-LIB2",
                                        "libraryType": "total_RNA_seq",
                                        "ngsLibraryConcentration": 0.005
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}""".lstrip()


def test_read_generic_custom_fields_sheet_header(tsv_sheet_generic_header):
    sheet = io_tsv.read_generic_tsv_sheet(tsv_sheet_generic_header)
    assert EXPECTED_GENERIC_SHEET_JSON_HEADER == json.dumps(
        sheet.json_data, indent='    ')


def test_read_generic_custom_fields_sheet_no_header(tsv_sheet_generic_no_header):
    with pytest.raises(io_tsv.TSVSheetException) as e_info:
        io_tsv.read_generic_tsv_json_data(tsv_sheet_generic_no_header)
    EXPECTED = (
        'Unexpected column seen in header row of body: ncbiTaxon, '
        'ngsLibraryConcentration, testSampleConcentration, uberonCellSource')
    assert EXPECTED == str(e_info.value)


def test_read_generic_sheet_custom_fields_json_header(tsv_sheet_generic_header):
    sheet_struc = io_tsv.read_generic_tsv_json_data(tsv_sheet_generic_header)
    assert EXPECTED_GENERIC_SHEET_JSON_HEADER == json.dumps(
        sheet_struc, indent='    ')


def test_read_generic_sheet_custom_fields_json_no_header(tsv_sheet_generic_no_header):
    with pytest.raises(io_tsv.TSVSheetException) as e_info:
        io_tsv.read_generic_tsv_json_data(tsv_sheet_generic_no_header)
    EXPECTED = (
        'Unexpected column seen in header row of body: ncbiTaxon, '
        'ngsLibraryConcentration, testSampleConcentration, uberonCellSource')
    assert EXPECTED == str(e_info.value)



# -*- coding: utf-8 -*-
"""Shortcuts for generic sample sheets
"""

from collections import OrderedDict

from .base import (ShortcutMixin)
from .base import (ShortcutSampleSheet)

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


class GenericSampleSheet(ShortcutSampleSheet):
    """Shortcut for "rare disease" view on bio-medical sample sheets"""

    def __init__(self, sheet):
        super().__init__(sheet)
        #: Generic wrapper BioEntity objects
        self.bio_entities = OrderedDict(self._build_bio_entities())
        #: List of all NGS libraries
        self.all_ngs_libraries = []
        #: List of the primary NGS libraries for each bio sample
        self.primary_ngs_libraries = []
        # Build the shortcuts
        self._build_shortcuts()

    def _build_bio_entities(self):
        """Build GenericBioEntity objects for ``self.sheet``"""
        for name, bio_entity in self.sheet.bio_entities.items():
            yield name, GenericBioEntity(self, bio_entity)

    def _build_shortcuts(self):
        # Build self.{all,primary}_ngs_libraries
        for bio_entity in self.bio_entities.values():
            for bio_sample in bio_entity.bio_samples.values():
                primary = True
                for test_sample in bio_sample.test_samples.values():
                    for ngs_library in (
                            test_sample.ngs_libraries.values()):
                        if ngs_library.disabled:
                            continue
                        self.all_ngs_libraries.append(ngs_library)
                        if primary:
                            self.primary_ngs_libraries.append(ngs_library)
                            primary = False


class GenericNGSLibrary(ShortcutMixin):
    """Shortcut wrapper for NGS library in generic sample sheet"""

    def __init__(self, shortcut_test_sample, ngs_library):
        #: Parent GenericTestSample
        self.test_sample = shortcut_test_sample
        #: Parent object; for ShortcutMixin
        self.parent = shortcut_test_sample
        #: Wrapped object
        self.wrapped = ngs_library
        #: Wrapped NGSLibrary
        self.ngs_library = ngs_library


class GenericTestSample(ShortcutMixin):
    """Shortcut wrapper for test sample in generic sample sheet"""

    #: Type to use for creating bio samples
    ngs_library_class = GenericNGSLibrary

    def __init__(self, shortcut_bio_sample, test_sample):
        #: Parent GenericBioSample
        self.bio_sample = shortcut_bio_sample
        #: Parent object; for ShortcutMixin
        self.parent = shortcut_bio_sample
        #: Wrapped object
        self.wrapped = test_sample
        #: Wrapped TestSample
        self.test_sample = test_sample
        #: Shortcut to NGSLibrary objects
        self.ngs_libraries = OrderedDict(self._build_ngs_libraries())

    def _build_ngs_libraries(self):
        """BuildGenericNGSLibrary objects from
        ``self.test_sample.ngs_libraries"""
        for name, ngs_library in self.test_sample.ngs_libraries.items():
            yield name, self.__class__.ngs_library_class(self, ngs_library)


class GenericBioSample(ShortcutMixin):
    """Shortcut wrapper for bio sample in generic sample sheet"""

    #: Type to use for creating bio samples
    test_sample_class = GenericTestSample

    def __init__(self, shortcut_bio_entity, bio_sample):
        #: Parent GenericBioEntity
        self.bio_entity = shortcut_bio_entity
        #: Parent object; for ShortcutMixin
        self.parent = shortcut_bio_entity
        #: Wrapped object
        self.wrapped = bio_sample
        #: Wrapped BioSample
        self.bio_sample = bio_sample
        #: Shortcut BioSample objects
        self.test_samples = OrderedDict(self._build_test_samples())

    def _build_test_samples(self):
        """Build GenericTestSample objects for ``self.bio_sample.test_samples``
        """
        for name, test_sample in self.bio_sample.test_samples.items():
            yield name, self.__class__.test_sample_class(self, test_sample)


class GenericBioEntity(ShortcutMixin):
    """Shortcut wrapper for bio entity in generic sample sheet"""

    #: Type to use for creating bio samples
    bio_sample_class = GenericBioSample

    def __init__(self, shortcut_sheet, bio_entity):
        #: Parent GenericSampleSheet
        self.sheet = shortcut_sheet
        #: Parent object; for ShortcutMixin
        self.parent = None
        #: Wrapped object; for ShortcutMixin
        self.wrapped = bio_entity
        #: Wrapped BioEntity
        self.bio_entity = bio_entity
        #: Shortcut BioSample objects
        self.bio_samples = OrderedDict(self._build_bio_samples())

    def _build_bio_samples(self):
        """Build GenericBioSample objects for ``self.bio_entity.bio_samples``
        """
        for name, bio_sample in self.bio_entity.bio_samples.items():
            yield name, self.__class__.bio_sample_class(self, bio_sample)



# -*- coding: utf-8 -*-
"""Code for reading and writing BioMed Sheets from/to file-like objects

Optionally, objects can be validated before reading/writing.
"""

from collections import OrderedDict
import json

from . import models
from .naming import DEFAULT_NAME_GENERATOR

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


def json_loads_ordered(s):
    """Helper function to load JSON using OrderedDict for object pairs"""
    return json.loads(s, object_pairs_hook=OrderedDict)


class SheetSchema:
    """Programmatic access to the bundled BioMed Schema JSON file"""

    @classmethod
    def load_from_string(self, s):
        """Load schema from JSON string ``s``"""
        return SheetSchema(json_loads_ordered(s))

    def __init__(self, parsed_json):
        #: parsed JSON that is wrapped by schema
        self.parsed_json = parsed_json


class ExtraInfoBuilder:
    """Helper class for converting from "extraInfo" value to Python value
    """

    def __init__(self, definition):
        #: Dispatch table for different types, mostly Python type constructors
        self.builders = {
            'boolean': bool,
            'string': str,
            'pattern': str,
            'integer': int,
            'number': float,
            'enum': str,
            'object': lambda x: x,  # identity
            'array': self._build_array,
        }
        #: Definition from JSON file
        self.definition = definition

    def build(self, value):
        return self.builders[self.definition['type']](value)

    def _build_array(self, value):
        """Handle array entry type"""
        func = self.builders[self.definition['entry']]
        return list(map(func, value))


class SheetBuilder:
    """Helper class to construct ``models.Sheet`` from JSON sheet data

    This is implemented as a class instead of a function so state can be
    stored in attributes and be used implicitely in the methods than
    having to be passed explicitely in function arguments
    """

    def __init__(self, json_data):
        #: Validated BioMed Sheet data
        self.json_data = json_data

    def run(self, dict_type=OrderedDict,
            name_generator=DEFAULT_NAME_GENERATOR):
        return models.Sheet(
            identifier=self.json_data.get('identifier', ''),
            title=self.json_data.get('title', ''),
            description=self.json_data.get('description', ''),
            bio_entities=self._build_bio_entities(
                extra_infos_defs=self.json_data.get(
                    'extraInfoDefs', dict_type()),
                bio_entities_json=self.json_data.get(
                    'bioEntities', dict_type()),
                dict_type=dict_type,
                name_generator=name_generator),
            json_data=self.json_data,
            dict_type=dict_type,
            name_generator=name_generator)

    def _build_bio_entities(
            self, extra_infos_defs, bio_entities_json, dict_type,
            name_generator):
        """Build BioEntity list
        """
        for secondary_id, value in bio_entities_json.items():
            bio_entity = models.BioEntity(
                pk=value['pk'],
                disabled=value.get('disabled', False),
                secondary_id=secondary_id,
                extra_ids=value.get('extraIds', []),
                extra_infos=dict_type(self._build_extra_infos(
                    extra_infos=value.get('extraInfo', dict_type()),
                    extra_infos_defs=extra_infos_defs.get(
                        'bioEntity', dict_type()),
                    dict_type=dict_type,
                    name_generator=name_generator)),
                bio_samples=dict_type(self._build_bio_samples(
                    extra_infos_defs=extra_infos_defs,
                    bio_samples_json=value.get('bioSamples', dict_type()),
                    dict_type=dict_type,
                    name_generator=name_generator)),
                dict_type=dict_type,
                name_generator=name_generator)
            yield (secondary_id, bio_entity)

    @classmethod
    def _build_extra_infos(
            cls, extra_infos_defs, extra_infos, dict_type, name_generator):
        """Interpret extraInfo stuff, including type conversion"""
        for key, value in extra_infos.items():
            yield (key, ExtraInfoBuilder(extra_infos_defs[key]).build(value))

    def _build_bio_samples(
            self, extra_infos_defs, bio_samples_json, dict_type,
            name_generator):
        """Build models.BioSample object, root JSON is required for attribute
        description
        """
        for secondary_id, value in bio_samples_json.items():
            bio_sample = models.BioSample(
                pk=value['pk'],
                disabled=value.get('disabled', False),
                secondary_id=secondary_id,
                extra_ids=value.get('extraIds', []),
                extra_infos=dict_type(self._build_extra_infos(
                    extra_infos=value.get('extraInfo', dict_type()),
                    extra_infos_defs=extra_infos_defs.get(
                        'bioSample', dict_type()),
                    dict_type=dict_type,
                    name_generator=name_generator)),
                test_samples=dict_type(self._build_test_samples(
                    extra_infos_defs, value.get('testSamples', dict_type()),
                    dict_type, name_generator)),
                dict_type=dict_type,
                name_generator=name_generator)
            yield (secondary_id, bio_sample)

    def _build_test_samples(
            self, extra_infos_defs, test_samples_json, dict_type,
            name_generator):
        """Build models.TestSample object
        """
        for secondary_id, value in test_samples_json.items():
            test_sample = models.TestSample(
                pk=value['pk'],
                disabled=value.get('disabled', False),
                secondary_id=secondary_id,
                extra_ids=value.get('extraIds', []),
                extra_infos=dict_type(self._build_extra_infos(
                    extra_infos=value.get('extraInfo', dict_type()),
                    extra_infos_defs=extra_infos_defs.get(
                        'testSample', dict_type()),
                    dict_type=dict_type,
                    name_generator=name_generator)),
                ngs_libraries=dict_type(self._build_ngs_libraries(
                    extra_infos_defs=extra_infos_defs,
                    ngs_libraries_json=value.get(
                        'ngsLibraries', dict_type()),
                    dict_type=dict_type,
                    name_generator=name_generator)),
                dict_type=dict_type,
                name_generator=name_generator)
            yield (secondary_id, test_sample)

    def _build_ngs_libraries(
            self, extra_infos_defs, ngs_libraries_json, dict_type,
            name_generator):
        """Build models.NGSLibrary objects
        """
        for secondary_id, value in ngs_libraries_json.items():
            ngs_library = models.NGSLibrary(
                pk=value['pk'],
                disabled=value.get('disabled', False),
                secondary_id=secondary_id,
                extra_ids=value.get('extraIds', []),
                extra_infos=dict_type(self._build_extra_infos(
                    extra_infos=value.get('extraInfo', dict_type()),
                    extra_infos_defs=extra_infos_defs.get(
                        'ngsLibrary', dict_type()),
                    dict_type=dict_type,
                    name_generator=name_generator)),
                dict_type=dict_type,
                name_generator=name_generator)
            yield (secondary_id, ngs_library)



# -*- coding: utf-8 -*-
"""Tests for loading the compressed cancer TSV file
"""

import io
import json
import textwrap

import pytest

from biomedsheets import io_tsv

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


@pytest.fixture
def tsv_sheet_cancer_header():
    f = io.StringIO(textwrap.dedent("""
    [Metadata]
    schema\ttumor_matched
    schema_version\tv1
    title\tExample matched cancer tumor/normal study
    description\tThe study has two patients, P001 has one tumor sample, P002 has two

    [Data]
    patientName\tsampleName\tisTumor\tlibraryType\tfolderName
    P001\tN1\tN\tWES\tP001-N1-DNA1-WES1
    P001\tT1\tY\tWES\tP001-T1-DNA1-WES1
    P001\tT1\tY\tmRNA_seq\tP001-T1-RNA1-mRNAseq1
    P002\tN1\tN\tWES\tP001-N1-DNA1-WES1
    P002\tT1\tY\tWES\tP001-T1-DNA1-WES1
    P002\tT1\tY\tWES\tP001-T1-RNA1-RNAseq1
    P002\tT2\tY\tWES\tP001-T2-DNA1-WES1
    P002\tT2\tY\tmRNA_seq\tP001-T2-RNA1-mRNAseq1
    """.lstrip()))
    return f


@pytest.fixture
def tsv_sheet_cancer_no_header():
    """Tumor TSV sheet without header"""
    f = io.StringIO(textwrap.dedent("""
    patientName\tsampleName\tisTumor\tlibraryType\tfolderName
    P001\tN1\tN\tWES\tP001-N1-DNA1-WES1
    P001\tT1\tY\tWES\tP001-T1-DNA1-WES1
    P001\tT1\tY\tmRNA_seq\tP001-T1-RNA1-mRNAseq1
    P002\tN1\tN\tWES\tP001-N1-DNA1-WES1
    P002\tT1\tY\tWES\tP001-T1-DNA1-WES1
    P002\tT1\tY\tWES\tP001-T1-RNA1-RNAseq1
    P002\tT2\tY\tWES\tP001-T2-DNA1-WES1
    P002\tT2\tY\tmRNA_seq\tP001-T2-RNA1-mRNAseq1
    """.lstrip()))
    return f


# Expected value for the cancer sheet JSON with header
EXPECTED_CANCER_SHEET_JSON_HEADER = r"""
{
    "identifier": "file://<unknown>",
    "title": "Example matched cancer tumor/normal study",
    "description": "The study has two patients, P001 has one tumor sample, P002 has two",
    "extraInfoDefs": {
        "bioEntity": {
            "ncbiTaxon": {
                "docs": "Reference to NCBI taxonomy",
                "key": "taxon",
                "type": "string",
                "pattern": "^NCBITaxon_[1-9][0-9]*$"
            }
        },
        "bioSample": {
            "isTumor": {
                "docs": "Boolean flag for distinguishing tumor/normal samples",
                "key": "isTumor",
                "type": "boolean"
            }
        },
        "testSample": {
            "extractionType": {
                "docs": "Describes extracted",
                "key": "extractionType",
                "type": "enum",
                "choices": [
                    "DNA",
                    "RNA",
                    "other"
                ]
            }
        },
        "ngsLibrary": {
            "seqPlatform": {
                "docs": "Sequencing platform used",
                "key": "kitName",
                "type": "enum",
                "choices": [
                    "Illumina",
                    "PacBio",
                    "other"
                ]
            },
            "libraryType": {
                "docs": "Rough classificiation of the library type",
                "key": "libraryType",
                "type": "enum",
                "choices": [
                    "Panel-seq",
                    "WES",
                    "WGS",
                    "mRNA-seq",
                    "tRNA-seq",
                    "other"
                ]
            },
            "folderName": {
                "docs": "Name of folder with FASTQ files",
                "key": "folderName",
                "type": "string"
            }
        }
    },
    "bioEntities": {
        "P001": {
            "pk": 1,
            "extraInfo": {
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {
                "N1": {
                    "pk": 2,
                    "extraInfo": {
                        "isTumor": false
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 3,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 4,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-N1-DNA1-WES1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        }
                    }
                },
                "T1": {
                    "pk": 5,
                    "extraInfo": {
                        "isTumor": true
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 6,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 7,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-DNA1-WES1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        },
                        "RNA1": {
                            "pk": 8,
                            "extraInfo": {
                                "extractionType": "RNA"
                            },
                            "ngsLibraries": {
                                "mRNA_seq1": {
                                    "pk": 9,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-RNA1-mRNAseq1",
                                        "libraryType": "mRNA_seq"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        "P002": {
            "pk": 10,
            "extraInfo": {
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {
                "N1": {
                    "pk": 11,
                    "extraInfo": {
                        "isTumor": false
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 12,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 13,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-N1-DNA1-WES1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        }
                    }
                },
                "T1": {
                    "pk": 14,
                    "extraInfo": {
                        "isTumor": true
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 15,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 16,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-DNA1-WES1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        },
                        "DNA2": {
                            "pk": 17,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES2": {
                                    "pk": 18,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-RNA1-RNAseq1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        }
                    }
                },
                "T2": {
                    "pk": 19,
                    "extraInfo": {
                        "isTumor": true
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 20,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 21,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T2-DNA1-WES1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        },
                        "RNA1": {
                            "pk": 22,
                            "extraInfo": {
                                "extractionType": "RNA"
                            },
                            "ngsLibraries": {
                                "mRNA_seq1": {
                                    "pk": 23,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T2-RNA1-mRNAseq1",
                                        "libraryType": "mRNA_seq"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}""".lstrip()

# Expected value for the cancer sheet JSON without header
EXPECTED_CANCER_SHEET_JSON_NO_HEADER = r"""
{
    "identifier": "file://<unknown>",
    "title": "Cancer Sample Sheet",
    "description": "Sample Sheet constructed from cancer matched samples compact TSV file",
    "extraInfoDefs": {
        "bioEntity": {
            "ncbiTaxon": {
                "docs": "Reference to NCBI taxonomy",
                "key": "taxon",
                "type": "string",
                "pattern": "^NCBITaxon_[1-9][0-9]*$"
            }
        },
        "bioSample": {
            "isTumor": {
                "docs": "Boolean flag for distinguishing tumor/normal samples",
                "key": "isTumor",
                "type": "boolean"
            }
        },
        "testSample": {
            "extractionType": {
                "docs": "Describes extracted",
                "key": "extractionType",
                "type": "enum",
                "choices": [
                    "DNA",
                    "RNA",
                    "other"
                ]
            }
        },
        "ngsLibrary": {
            "seqPlatform": {
                "docs": "Sequencing platform used",
                "key": "kitName",
                "type": "enum",
                "choices": [
                    "Illumina",
                    "PacBio",
                    "other"
                ]
            },
            "libraryType": {
                "docs": "Rough classificiation of the library type",
                "key": "libraryType",
                "type": "enum",
                "choices": [
                    "Panel-seq",
                    "WES",
                    "WGS",
                    "mRNA-seq",
                    "tRNA-seq",
                    "other"
                ]
            },
            "folderName": {
                "docs": "Name of folder with FASTQ files",
                "key": "folderName",
                "type": "string"
            }
        }
    },
    "bioEntities": {
        "P001": {
            "pk": 1,
            "extraInfo": {
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {
                "N1": {
                    "pk": 2,
                    "extraInfo": {
                        "isTumor": false
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 3,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 4,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-N1-DNA1-WES1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        }
                    }
                },
                "T1": {
                    "pk": 5,
                    "extraInfo": {
                        "isTumor": true
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 6,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 7,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-DNA1-WES1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        },
                        "RNA1": {
                            "pk": 8,
                            "extraInfo": {
                                "extractionType": "RNA"
                            },
                            "ngsLibraries": {
                                "mRNA_seq1": {
                                    "pk": 9,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-RNA1-mRNAseq1",
                                        "libraryType": "mRNA_seq"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        "P002": {
            "pk": 10,
            "extraInfo": {
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {
                "N1": {
                    "pk": 11,
                    "extraInfo": {
                        "isTumor": false
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 12,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 13,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-N1-DNA1-WES1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        }
                    }
                },
                "T1": {
                    "pk": 14,
                    "extraInfo": {
                        "isTumor": true
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 15,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 16,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-DNA1-WES1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        },
                        "DNA2": {
                            "pk": 17,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES2": {
                                    "pk": 18,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-RNA1-RNAseq1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        }
                    }
                },
                "T2": {
                    "pk": 19,
                    "extraInfo": {
                        "isTumor": true
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 20,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 21,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T2-DNA1-WES1",
                                        "libraryType": "WES"
                                    }
                                }
                            }
                        },
                        "RNA1": {
                            "pk": 22,
                            "extraInfo": {
                                "extractionType": "RNA"
                            },
                            "ngsLibraries": {
                                "mRNA_seq1": {
                                    "pk": 23,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T2-RNA1-mRNAseq1",
                                        "libraryType": "mRNA_seq"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}""".lstrip()


def test_read_cancer_sheet_header(tsv_sheet_cancer_header):
    sheet = io_tsv.read_cancer_tsv_sheet(tsv_sheet_cancer_header)
    assert EXPECTED_CANCER_SHEET_JSON_HEADER == json.dumps(
        sheet.json_data, indent='    ')


def test_read_cancer_sheet_no_header(tsv_sheet_cancer_no_header):
    sheet = io_tsv.read_cancer_tsv_sheet(tsv_sheet_cancer_no_header)
    assert EXPECTED_CANCER_SHEET_JSON_NO_HEADER == json.dumps(
        sheet.json_data, indent='    ')


def test_read_tumor_json_header(tsv_sheet_cancer_header):
    sheet_struc = io_tsv.read_cancer_tsv_json_data(tsv_sheet_cancer_header)
    assert EXPECTED_CANCER_SHEET_JSON_HEADER == json.dumps(
        sheet_struc, indent='    ')


def test_read_tumor_json_no_header(tsv_sheet_cancer_no_header):
    sheet_struc = io_tsv.read_cancer_tsv_json_data(tsv_sheet_cancer_no_header)
    assert EXPECTED_CANCER_SHEET_JSON_NO_HEADER == json.dumps(
        sheet_struc, indent='    ')



# -*- coding: utf-8 -*-
"""Python classes for representing the generic part of BioMedical sheets
"""

from collections import OrderedDict

from .naming import DEFAULT_NAME_GENERATOR

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


#: Key for storing disabled flag for entities
KEY_DISABLED = 'disabled'
# TODO: add this as a main property

#: Key for selecting an ``NGSLibrary`` object
KEY_NGS_LIBRARY = 'ngs_library'

#: Frozen set of valid ``TestSample`` child types
TEST_SAMPLE_CHILDREN = frozenset((KEY_NGS_LIBRARY,))


class BioMedSheetsBaseException(Exception):
    """Base exception for module"""


class AmbiguousSecondaryIdException(BioMedSheetsBaseException):
    """Raised on duplicate secondary IDs"""


class SheetPathCrawlingException(Exception):
    """Raised on problems crawling the sample sheet"""


class InvalidSecondaryIDException(SheetPathCrawlingException):
    """Raised when the secondary ID was invalid"""


class SecondaryIDNotFoundException(SheetPathCrawlingException):
    """Raised when a secondary id could not be found during crawling"""


class CrawlMixin:
    """Mixin that provides the ``crawl()`` function

    Also provides helpers for merging "sub_entries" dicts
    """

    def crawl(self, name, sep='-'):
        """Crawl through sheet based on the path by secondary id
        """
        if sep in name:
            next, rest = name.split(sep, 1)
        else:
            next, rest = name, None
        if next not in self.sub_entries:
            raise SecondaryIDNotFoundException(
                'Could not find sub entry with secondary ID {}'.format(
                    next))
        if rest:
            return self.sub_entries[next].crawl(rest)
        else:
            return self.sub_entries[next]

    def _merge_sub_entries(self, *dicts):
        # Check for conflicts in secondary ids
        duplicates = set()
        for d1 in dicts:
            for d2 in dicts:
                if d1 is not d2:
                    dupes = set(d1.keys()) & set(d2.keys())
                    duplicates = duplicates | dupes
        if len(duplicates) > 0:
            raise AmbiguousSecondaryIdException(
                'Ambiguous secondary IDs: {}'.format(duplicates))
        # Build result
        result = {}
        for d in dicts:
            result.update(d)
        return result


class Sheet(CrawlMixin):
    """Container for multiple :class:`BioEntity` objects
    """

    def __init__(self, identifier, title, json_data, description=None,
                 bio_entities=None, extra_infos=None, dict_type=OrderedDict,
                 name_generator=DEFAULT_NAME_GENERATOR):
        #: Identifier URI of the sheet, cannot be changed after construction
        self.identifier = identifier
        #: Title of the sheet, can be changed after construction
        self.title = title
        #: The underlying data from JSON
        self.json_data = json_data
        #: Description of the sheet
        self.description = description
        #: Extra info, ``dict``-like object
        self.extra_infos = dict_type(extra_infos or [])
        #: List of ``BioEntity`` objects described in the sheet
        self.bio_entities = dict_type(bio_entities or [])
        #: Create ``sub_entries`` shortcut for ``crawl()``
        self.sub_entries = self.bio_entities
        #: Name generator used in the sheet
        self.name_generator = name_generator

    def __repr__(self):
        return 'Sheet({})'.format(', '.join(map(str, [
            self.identifier, self.title, self.json_data, self.description,
            self.bio_entities, self.extra_infos, self.sub_entries])))

    def __str__(self):
        return repr(self)


class SheetEntry:
    """Base class for the different ``Sheet`` entries

    Pulls up the common properties of primary key, secondary ID and additional
    properties dict
    """

    def __init__(self, pk, disabled, secondary_id, extra_ids=None,
                 extra_infos=None, dict_type=OrderedDict,
                 name_generator=DEFAULT_NAME_GENERATOR):
        #: Primary key of the bio entity, globally unique
        self.pk = pk
        #: Flag for explicit disabling of objects
        self.disabled = disabled
        #: ``str`` with secondary id fragment of the bio entity, unique in the
        #: sheet
        self.secondary_id = secondary_id
        #: Extra IDs
        self.extra_ids = list(extra_ids or [])
        #: Extra info, ``dict``-like object
        self.extra_infos = dict_type(extra_infos or [])
        #: Name generator to use
        self.name_generator = name_generator

    @property
    def full_secondary_id(self):
        """Return full path from BioEntity to this entry"""
        raise NotImplementedError('Override me!')

    @property
    def name(self):
        """Abstract function for returning name"""
        return self.name_generator(self)

    @property
    def enabled(self):
        """Inverse of ``self.enabled``"""
        return not self.disabled


class BioEntity(SheetEntry, CrawlMixin):
    """Represent one biological specimen
    """

    def __init__(self, pk, disabled, secondary_id, extra_ids=None,
                 extra_infos=None, bio_samples=None, dict_type=OrderedDict,
                 name_generator=DEFAULT_NAME_GENERATOR):
        super().__init__(
            pk, disabled, secondary_id, extra_ids, extra_infos, dict_type,
            name_generator)
        #: List of ``BioSample`` objects described for the ``BioEntity``
        self.bio_samples = dict_type(bio_samples or [])
        # Assign owner pointer in bio samples to self
        for bio_sample in self.bio_samples.values():
            bio_sample.bio_entity = self
        #: Create ``sub_entries`` shortcut for ``crawl()``
        self.sub_entries = self.bio_samples

    @property
    def full_secondary_id(self):
        return self.secondary_id

    def __repr__(self):
        return 'BioEntity({})'.format(', '.join(map(str, [
            self.pk, self.disabled, self.secondary_id, self.extra_ids,
            self.extra_infos, self.bio_samples])))

    def __str__(self):
        return repr(self)


class BioSample(SheetEntry, CrawlMixin):
    """Represent one sample taken from a biological entity/specimen
    """

    def __init__(self, pk, disabled, secondary_id, extra_ids=None,
                 extra_infos=None, test_samples=None, dict_type=OrderedDict,
                 bio_entity=None, name_generator=DEFAULT_NAME_GENERATOR):
        super().__init__(
            pk, disabled, secondary_id, extra_ids, extra_infos, dict_type,
            name_generator)
        #: Containing BioEntity
        self.bio_entity = bio_entity
        #: List of ``TestSample`` objects described for the ``BioSample``
        self.test_samples = dict_type(test_samples or [])
        # Assign owner pointer in test samples to self
        for test_sample in self.test_samples.values():
            test_sample.bio_sample = self
        #: Create ``sub_entries`` shortcut for ``crawl()``
        self.sub_entries = self.test_samples

    @property
    def full_secondary_id(self):
        return '-'.join((self.bio_entity.full_secondary_id, self.secondary_id))

    def __repr__(self):
        return 'BioSample({})'.format(', '.join(map(str, [
            self.pk, self.disabled, self.secondary_id, self.extra_ids,
            self.extra_infos, self.test_samples])))

    def __str__(self):
        return repr(self)


class TestSample(SheetEntry, CrawlMixin):
    """Represent a technical sample from biological sample, e.g., DNA or RNA
    """

    def __init__(self, pk, disabled, secondary_id, extra_ids=None,
                 extra_infos=None, ngs_libraries=None, dict_type=OrderedDict,
                 bio_sample=None,
                 name_generator=DEFAULT_NAME_GENERATOR):
        super().__init__(
            pk, disabled, secondary_id, extra_ids, extra_infos, dict_type,
            name_generator)
        #: Containing BioSample
        self.bio_sample = bio_sample
        #: List of ``NGSLibrary`` objects described for the ``TestSample``
        self.ngs_libraries = dict_type(ngs_libraries or [])
        # Assign owner pointer in NGS libraries to self
        for ngs_library in self.ngs_libraries.values():
            ngs_library.test_sample = self
        # ``sub_entries`` shortcut, check for duplicates
        self.sub_entries = self._merge_sub_entries(self.ngs_libraries)

    @property
    def full_secondary_id(self):
        return '-'.join((self.bio_sample.full_secondary_id, self.secondary_id))

    def __repr__(self):
        return 'TestSample({})'.format(', '.join(map(str, [
            self.pk, self.disabled, self.secondary_id, self.extra_ids,
            self.extra_infos, self.ngs_libraries])))

    def __str__(self):
        return repr(self)


class NGSLibrary(SheetEntry):
    """Represent one NGSLibrary generated from a test sample
    """

    def __init__(self, pk, disabled, secondary_id, extra_ids=None,
                 extra_infos=None, dict_type=OrderedDict, test_sample=None,
                 name_generator=DEFAULT_NAME_GENERATOR):
        super().__init__(
            pk, disabled, secondary_id, extra_ids, extra_infos, dict_type,
            name_generator)
        #: Owning TestSample
        self.test_sample = test_sample

    @property
    def full_secondary_id(self):
        return '-'.join((
            self.test_sample.full_secondary_id, self.secondary_id))

    def __repr__(self):
        return 'NGSLibrary({})'.format(', '.join(map(str, [
            self.pk, self.disabled, self.secondary_id, self.extra_ids,
            self.extra_infos])))

    def __str__(self):
        return repr(self)



# -*- coding: utf-8 -*
"""Helper code for name generation"""

import re


#: PK padding length
DEFAULT_PK_PADDING_LENGTH = 6

#: Default name pattern
NAME_PATTERN_DEFAULT = "{full_secondary_id}-{pk}"

#: Name pattern with secondary id only
NAME_PATTERN_SECONDARY_ID = "{full_secondary_id}"

#: Naming scheme "{secondary_id}-{pk}"
NAMING_SECONDARY_ID_PK = 'secondary_id_pk'

#: Naming scheme "{secondary_id}"
NAMING_ONLY_SECONDARY_ID = 'only_secondary_id'

#: Default naming scheme
NAMING_DEFAULT = NAMING_SECONDARY_ID_PK

#: Valid naming schemes
NAMING_SCHEMES = (NAMING_SECONDARY_ID_PK, NAMING_ONLY_SECONDARY_ID)

#: Inverse mapping from name pattern to RE for going from name to secondary id
#: and/or pk
INVERSE_PATTERN_MAP = {
    NAME_PATTERN_DEFAULT: r'^(?P<secondary_id>.*)-(?P<pk>[0-9]+)$',
    NAME_PATTERN_SECONDARY_ID: r'^(?P<secondary_id>.*)$',
}


class NameGenerator:
    """Base class for name generators for BioMed Sheet entities"""

    def __call__(self, obj):
        """Return generated name"""
        raise NotImplementedError(  # pragma: no cover
            'Abstract method called!')


class PatternNameGenerator(NameGenerator):
    """Name generation based on string format patterns"""

    def __init__(self, pattern, pk_padding_length=DEFAULT_PK_PADDING_LENGTH):
        #: The pattern to generate the name for
        self.pattern = pattern
        #: The default padding length
        self.pk_padding_length = pk_padding_length
        #: RE for mapping from name to secondary id or PK
        if pattern not in INVERSE_PATTERN_MAP:
            raise ValueError(  # pragma: no cover
                'Cannot map back from pattern {}'.format(pattern))
        self.inverse_name_re = INVERSE_PATTERN_MAP[self.pattern]

    def inverse(self, name, component='secondary_id'):
        """Map from name to secondary_id/pk (given in ``component``)"""
        assert component in ('secondary_id', 'pk')
        match = re.match(self.inverse_name_re, name)
        if not match:
            raise ValueError('Could not match {}'.format(name))
        return match.group(component)

    def __call__(self, obj):
        """Return generated name"""
        padded_pk = str(obj.pk).rjust(self.pk_padding_length, '0')
        return self.pattern.format(
            full_secondary_id=obj.full_secondary_id, pk=padded_pk)


#: Default name generator is a :py:class:`PatternNameGenerator` with
#: :py:data:`NAME_PATTERN_DEFAULT` as the value.
DEFAULT_NAME_GENERATOR = PatternNameGenerator(NAME_PATTERN_DEFAULT)


def name_generator_for_scheme(scheme):
    """Return ``PatternNameGenerator`` for the given naming ``scheme``"""
    if scheme not in NAMING_SCHEMES:
        raise ValueError(  # pragma: no cover
            'No known naming scheme {}'.format(scheme))
    mapping = {
        NAMING_SECONDARY_ID_PK: PatternNameGenerator(NAME_PATTERN_DEFAULT),
        NAMING_ONLY_SECONDARY_ID: PatternNameGenerator(
            NAME_PATTERN_SECONDARY_ID),
    }
    return mapping[scheme]



# -*- coding: utf-8 -*
"""Code for convenient access to sample sheets"""

from .base import *  # noqa: F401,F403
from .generic import *  # noqa: F401,F403
from .cancer import *  # noqa: F401,F403
from .germline import *  # noqa: F401,F403

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'
